import { CharString } from './CharString';
import { ComplexDecimal } from './ComplexDecimal';
import { FunctionHandle } from './FunctionHandle';
import { type ElementType, MultiArray } from './MultiArray';
/**
 * AST (Abstract Syntax Tree).
 */
/**
 * Operator type.
 */
type TOperator = '+' | '-' | '.*' | '*' | './' | '/' | '.\\' | '\\' | '.^' | '^' | '.**' | '**' | '<' | '<=' | '==' | '>=' | '>' | '!=' | '~=' | '&' | '|' | '&&' | '||' | '=' | '+=' | '-=' | '*=' | '/=' | '\\=' | '^=' | '**=' | '.*=' | './=' | '.\\=' | '.^=' | '.**=' | '&=' | '|=' | '()' | '!' | '~' | '+_' | '-_' | '++_' | '--_' | ".'" | "'" | '_++' | '_--';
/**
 * Common primary node.
 */
interface PrimaryNode {
    type: string | number;
    parent?: any;
    index?: number;
    omitOut?: boolean;
    start?: {
        line: number;
        column: number;
    };
    stop?: {
        line: number;
        column: number;
    };
}
type NodeInput = NodeExpr | NodeList | NodeDeclaration | NodeIf;
/**
 * Expression node.
 */
type NodeExpr = ElementType | NodeIdentifier | NodeIndexExpr | NodeOperation | NodeRange | NodeIndirectRef | NodeReturnList | any;
/**
 * Reserved node.
 */
interface NodeReserved extends PrimaryNode {
}
/**
 * Literal node.
 */
interface NodeLiteral extends PrimaryNode {
}
/**
 * Name node.
 */
interface NodeIdentifier extends PrimaryNode {
    type: 'IDENT';
    id: string;
}
/**
 * Command word list node.
 */
interface NodeCmdWList extends PrimaryNode {
    type: 'CMDWLIST';
    id: string;
    args: CharString[];
    omitAns?: boolean;
}
/**
 * Expression and arguments node.
 */
interface NodeIndexExpr extends PrimaryNode {
    type: 'IDX';
    expr: NodeExpr;
    args: NodeExpr[];
    delim: '()' | '{}';
}
/**
 * Range node.
 */
interface NodeRange extends PrimaryNode {
    type: 'RANGE';
    start_: NodeExpr | null;
    stop_: NodeExpr | null;
    stride_: NodeExpr | null;
}
/**
 * Operation node.
 */
type NodeOperation = UnaryOperation | BinaryOperation;
/**
 * Unary operation node.
 */
type UnaryOperation = UnaryOperationL | UnaryOperationR;
/**
 * Right unary operation node.
 */
interface UnaryOperationR extends PrimaryNode {
    right: NodeExpr;
}
/**
 * Left unary operation node.
 */
interface UnaryOperationL extends PrimaryNode {
    left: NodeExpr;
    omitAns?: boolean;
}
/**
 * Binary operation.
 */
interface BinaryOperation extends PrimaryNode {
    left: NodeExpr;
    right: NodeExpr;
    omitAns?: boolean;
}
/**
 * List node
 */
interface NodeList extends PrimaryNode {
    type: 'LIST';
    list: NodeInput[];
}
interface NodeIndirectRef extends PrimaryNode {
    type: '.';
    obj: NodeExpr;
    field: (string | NodeExpr)[];
}
type ReturnSelector = (length: number, index: number) => any;
/**
 * Return list node
 */
interface NodeReturnList extends PrimaryNode {
    type: 'RETLIST';
    selector: ReturnSelector;
}
interface NodeFunction extends PrimaryNode {
    type: 'FCN';
    id: string;
    return: NodeInput[];
    parameter: NodeInput[];
    arguments: NodeInput[];
    statements: NodeInput[];
}
interface NodeArgumentValidation {
    name: NodeIdentifier;
    size: NodeInput[];
    class: NodeIdentifier | null;
    functions: NodeInput[];
    default: NodeExpr;
}
interface NodeArguments {
    attribute: NodeIdentifier | null;
    validation: NodeArgumentValidation[];
}
interface NodeDeclaration extends PrimaryNode {
    type: 'GLOBAL' | 'PERSIST';
    list: NodeExpr[];
}
interface NodeIf extends PrimaryNode {
    type: 'IF';
    expression: NodeExpr[];
    then: NodeList[];
    else: NodeList | null;
    omitAns?: boolean;
}
interface NodeElseIf {
    expression: NodeExpr;
    then: NodeList;
}
interface NodeElse {
    else: NodeList;
}
/**
 * AST (Abstract Syntax Tree) constructor methods.
 */
declare const nodeString: typeof CharString.create;
declare const nodeNumber: typeof ComplexDecimal.parse;
declare const firstRow: typeof MultiArray.firstRow;
declare const appendRow: typeof MultiArray.appendRow;
declare const emptyArray: typeof MultiArray.emptyArray;
/**
 * Create literal node.
 * @param type
 * @returns
 */
declare const nodeLiteral: (type: string) => NodeLiteral;
/**
 * Create name node.
 * @param nodeid
 * @returns
 */
declare const nodeIdentifier: (id: string) => NodeIdentifier;
/**
 * Create command word list node.
 * @param nodename
 * @param nodelist
 * @returns
 */
declare const nodeCmdWList: (nodename: NodeIdentifier, nodelist: NodeList) => NodeCmdWList;
/**
 * Create expression and arguments node.
 * @param nodeexpr
 * @param nodelist
 * @returns
 */
declare const nodeIndexExpr: (nodeexpr: NodeExpr, nodelist?: NodeList | null, delimiter?: "()" | "{}") => NodeIndexExpr;
/**
 * Create range node.
 * @param start_
 * @param stop_
 * @param stride_
 * @returns NodeRange.
 */
declare const nodeRange: (start_: NodeExpr, stop_: NodeExpr, stride_?: NodeExpr) => NodeRange;
/**
 * Create operator node.
 * @param op
 * @param data1
 * @param data2
 * @returns
 */
declare const nodeOp: (op: TOperator, data1: NodeExpr, data2?: NodeExpr) => NodeOperation;
/**
 * Create first element of list node.
 * @param node First element of list node.
 * @returns A NodeList.
 */
declare const nodeListFirst: (node?: NodeInput) => NodeList;
/**
 * Append node to list node.
 * @param lnode NodeList.
 * @param node Element to append to list.
 * @returns NodeList with element appended.
 */
declare const appendNodeList: (lnode: NodeList, node: NodeInput) => NodeList;
declare const nodeList: (list: NodeInput[]) => NodeList;
/**
 * Create first row of a MultiArray.
 * @param row
 * @returns
 */
declare const nodeFirstRow: (row?: NodeList | null, iscell?: boolean) => MultiArray;
/**
 * Append row to MultiArray.
 * @param M
 * @param row
 * @returns
 */
declare const nodeAppendRow: (M: MultiArray, row?: NodeList | null) => MultiArray;
declare const nodeIndirectRef: (left: NodeExpr, right: string | NodeExpr) => NodeIndirectRef;
/**
 * Creates NodeReturnList (multiple assignment)
 * @param selector Left side selector function.
 * @returns Return list node.
 */
declare const nodeReturnList: (selector: ReturnSelector) => NodeReturnList;
declare const nodeFunctionHandle: (id?: NodeIdentifier | null, parameter_list?: NodeList | null, expression?: NodeExpr) => FunctionHandle;
declare const nodeFunction: (id: NodeIdentifier, return_list: NodeList, parameter_list: NodeList, arguments_list: NodeList, statements_list: NodeList) => NodeFunction;
declare const nodeArgumentValidation: (name: NodeIdentifier, size: NodeList, cl: (NodeIdentifier | null) | undefined, functions: NodeList, dflt?: NodeExpr) => NodeArgumentValidation;
declare const nodeArguments: (attribute: NodeIdentifier | null, validationList: NodeList) => NodeArguments;
declare const nodeDeclarationFirst: (type: "GLOBAL" | "PERSIST") => NodeDeclaration;
declare const nodeAppendDeclaration: (node: NodeDeclaration, declaration: NodeExpr) => NodeDeclaration;
declare const nodeIfBegin: (expression: NodeExpr, then: NodeList) => NodeIf;
declare const nodeIfAppendElse: (nodeIf: NodeIf, nodeElse: NodeElse) => NodeIf;
declare const nodeIfAppendElseIf: (nodeIf: NodeIf, nodeElseIf: NodeElseIf) => NodeIf;
declare const nodeElseIf: (expression: NodeExpr, then: NodeList) => NodeElseIf;
declare const nodeElse: (elseStmt: NodeList) => NodeElse;
export type { TOperator, NodeInput, NodeExpr, NodeOperation, UnaryOperation, ReturnSelector };
export { PrimaryNode, NodeReserved, NodeLiteral, NodeIdentifier, NodeCmdWList, NodeIndexExpr, NodeRange, UnaryOperationR, UnaryOperationL, BinaryOperation, NodeList, NodeIndirectRef, NodeReturnList, NodeFunction, NodeArgumentValidation, NodeArguments, NodeDeclaration, NodeIf, NodeElseIf, NodeElse, nodeString, nodeNumber, firstRow, appendRow, emptyArray, nodeLiteral, nodeIdentifier, nodeCmdWList, nodeIndexExpr, nodeRange, nodeOp, nodeListFirst, appendNodeList, nodeList, nodeFirstRow, nodeAppendRow, nodeIndirectRef, nodeReturnList, nodeFunctionHandle, nodeFunction, nodeArgumentValidation, nodeArguments, nodeDeclarationFirst, nodeAppendDeclaration, nodeIfBegin, nodeIfAppendElse, nodeIfAppendElseIf, nodeElseIf, nodeElse, };
